#!/usr/bin/python

# Copyright 2009 Milos Jakubicek, Vojtech Kovar

# This file is part of SET.
#
# SET is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SET.  If not, see <http://www.gnu.org/licenses/>.


import signal
signal.signal(signal.SIGINT, signal.SIG_DFL) # Quit on Ctrl+C

import sys, locale
from PyQt4.QtCore import *
from PyQt4.QtGui import *
from PyQt4.QtSvg import *


class Node(QPoint):# represents a tree node in the graphical view

    def __init__(self, id, label, dep_id, dep_type, from_left, from_top):
        QPoint.__init__(self, from_left, from_top)
        self.id = int(id)
        self.label = unicode(label, locale.getpreferredencoding())
        self.dep_id = int(dep_id)
        self.dep_type = dep_type

class TreeViewerForm(QMainWindow):
    
    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        from TreeViewerUI import Ui_MainWindow
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)
        self.defaultFont = self.font()

    def hSliderSub(self):
        self.ui.hSlider.triggerAction(QSlider.SliderSingleStepSub)

    def hSliderAdd(self):
        self.ui.hSlider.triggerAction(QSlider.SliderSingleStepAdd)

    def vSliderSub(self):
        self.ui.vSlider.triggerAction(QSlider.SliderSingleStepSub)

    def vSliderAdd(self):
        self.ui.vSlider.triggerAction(QSlider.SliderSingleStepAdd)

    def scaleSliderSub(self):
        self.ui.scaleSlider.triggerAction(QSlider.SliderSingleStepSub)

    def scaleSliderAdd(self):
        self.ui.scaleSlider.triggerAction(QSlider.SliderSingleStepAdd)

    def firstTree(self):
        self.setTree(0)

    def lastTree(self):
        self.setTree(len(self.ui.treeview.trees) - 1)

    def nextTree(self):
        if self.ui.treeview.treeNumber != len(self.ui.treeview.trees) -1:
            self.setTree(self.ui.treeview.treeNumber + 1)

    def previousTree(self):
        if self.ui.treeview.treeNumber != 0:
            self.setTree(self.ui.treeview.treeNumber - 1)

    def setTree(self, treeNumber):
        self.ui.treeview.treeNumber = treeNumber
        self.ui.treeNumber.setValue(treeNumber)
        self.ui.treeview.compute_tree()
        if treeNumber == 0:
            self.ui.previousButton.setEnabled(False)
            self.ui.firstButton.setEnabled(False)
        else:
            self.ui.previousButton.setEnabled(True)
            self.ui.firstButton.setEnabled(True)
        if treeNumber == len(self.ui.treeview.trees) - 1:
            self.ui.nextButton.setEnabled(False)
            self.ui.lastButton.setEnabled(False)
        else:
            self.ui.nextButton.setEnabled(True)
            self.ui.lastButton.setEnabled(True)
        self.ui.treeview.update()
    
    def restoreDefaults(self):
        self.ui.treeview.y_space_quotient = 3
        self.ui.treeview.x_space_quotient = 5
        self.ui.treeview.scale = 1
        self.ui.treeview.changeFont(self.defaultFont)
        self.ui.hSlider.setValue(5)
        self.ui.vSlider.setValue(3)
        self.ui.scaleSlider.setValue(100)
        self.ui.treeview.update()

    def exportImage(self):
        imageFile = QFileDialog.getSaveFileName(self)
        if len(imageFile) > 0:
            if self.ui.imageTypes.currentText() == "SVG":
                self.ui.treeview.paintSVG(imageFile, self.ui.sentence.text())
            else:
                self.ui.treeview.paintImage(imageFile, str(self.ui.imageTypes.currentText()))

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Control:
            self.ui.treeview.ctrlPressed = True

    def keyReleaseEvent(self, event):
        if event.key() == Qt.Key_Control:
            self.ui.treeview.ctrlPressed = False

class TreeView(QWidget):

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        self.treeNumber = None
        self.nodes = None
        self.y_space_quotient = 3
        self.y_space = self.fontMetrics().height() * self.y_space_quotient
        self.x_space_quotient = 5
        self.x_space = self.fontMetrics().averageCharWidth() * self.x_space_quotient
        self.from_top = self.fontMetrics().height() * 1.5
        self.from_left = self.fontMetrics().averageCharWidth() * 5
        self.trees = None
        self.scale = 1
        self.width = 0
        self.height = 0
        self.ctrlPressed = False

    def changeXSpaceQuotient(self, quotient):
        self.x_space_quotient = quotient
        self.x_space = self.fontMetrics().averageCharWidth() * self.x_space_quotient
        self.compute_tree()
        self.update()

    def changeYSpaceQuotient(self, quotient):
        self.y_space_quotient = quotient
        self.y_space = self.fontMetrics().height() * self.y_space_quotient
        self.compute_tree()
        self.update()

    def chooseFont(self):
        font, ok = QFontDialog.getFont(self)
        if ok:
            self.changeFont(font)
 
    def changeFont(self, font):
        self.setFont(font)
        self.y_space = self.fontMetrics().height() * self.y_space_quotient
        self.x_space = self.fontMetrics().averageCharWidth() * self.x_space_quotient
        self.from_top = self.fontMetrics().height() * 1.5
        self.from_left = self.fontMetrics().averageCharWidth() * 5
        self.compute_tree()
   
    def changeScale(self, scale):
        if scale < 1:
            scale = 1
        self.scale = scale / 100.0
        self.update()

    def wheelEvent(self, event):
        if self.ctrlPressed:
            self.changeScale(self.scale * 100 + event.delta() / 10.0)
        else:
            event.ignore()

    def paintEvent(self, event):
        painter = QPainter()
        painter.begin(self)
        self.paint(painter)
        self.setMinimumSize(self.width * self.scale, self.height * self.scale)

    def paintSVG(self, fileName, sentence):
        generator = QSvgGenerator()
        generator.setFileName(fileName)
        generator.setSize(QSize(self.width, self.height))
        generator.setTitle("Syntactic tree")
        generator.setDescription("Syntactic tree for the input sentence '" + sentence + "', generated by Tree Viewer.");
        painter = QPainter()
        painter.begin(generator)
        self.paint(painter)

    def paintImage(self, fileName, imageType):
        image = QImage(self.width, self.height, QImage.Format_ARGB32)
        painter = QPainter()
        painter.begin(image)
        self.paint(painter)
        image.save(fileName, imageType)

    def printImage(self):
        printer = QPrinter()
        printDialog = QPrintDialog(printer, self.parent())
        if printDialog.exec_() == QDialog.Accepted:
            painter = QPainter()
            painter.begin(printer)
            self.paint(painter)
 
    def paint(self, painter):
        painter.scale(self.scale, self.scale)
        sentence = []
        maxX = 0
        maxY = 0

        for node in self.nodes:
            if node.label.startswith('<'):
                painter.setPen(QColor("blue"))
            else:
                painter.setPen(QColor("black"))
                sentence.append(node.label)
            painter.drawText(node.x() - painter.fontMetrics().width(node.label) / 2.0, node.y(), node.label)
            if node.dep_id > -1:
                if node.dep_id < 10000:
                    depNode = self.nodes[node.dep_id]
                else:
                    depNode = self.nodes[-1] # nasty hack
                if node.dep_type == 'p':
                    painter.setPen(QColor("blue"))
                else:
                    painter.setPen(QColor("black"))
                painter.drawLine(depNode.x(), depNode.y() + painter.fontMetrics().height() * 0.25, node.x(), node.y() - painter.fontMetrics().height() * 0.75)
            if maxX < node.x() + painter.fontMetrics().width(node.label) / 2.0:
                maxX = node.x() + painter.fontMetrics().width(node.label) / 2.0
            if maxY < node.y():
                maxY = node.y()
       
        self.width = maxX
        self.height = maxY
        self.emit(SIGNAL("treeChanged(QString)"), " ".join(filter(lambda word: word != "[root]", sentence)))
        painter.end()

    def compute_tree(self):
        # TODO divide the code
        nodes = []
        for line in self.trees[self.treeNumber]:
            id, label, dep_id, dep_type = line.split('\t')
            nodes.append(Node(id, label, dep_id, dep_type, self.from_left, self.from_top))
        # x coordinations
        x = self.from_left
        for index, node in enumerate(nodes):
            node.setX(x)
            if not node.label.startswith('<'):
                shift = self.fontMetrics().width(node.label) / 2.0
                if index + 1 < len(nodes):
                    shift += self.fontMetrics().width(nodes[index+1].label) / 2.0
                x += self.x_space + shift
        # y coordinations
        nodes_from_bottom = []
        y = self.from_top
        next_level = [-1]
        next_next_level = []
        while next_level:
            for id in next_level:
                for node in nodes:
                    if node.dep_id == id:
                        nodes_from_bottom.append(node)
                        node.setY(y)
                        next_next_level.append(node.id)
            next_level = next_next_level
            next_next_level = []
            y += self.y_space
        nodes_from_bottom.reverse()
        # x coordinations for phrasal tokens
        for phr_node in nodes_from_bottom:
            if not phr_node.label.startswith('<'): continue
            first = -1; last = -1
            for node in nodes:
                if node.dep_id == phr_node.id: # take into account
                    last = node.x()
                    if first < 0: first = node.x()
            if first > 0: phr_node.setX((last + first) / 2.0)
            for node in nodes: # check if not violating another token
                if phr_node.y() == node.y() and not phr_node == node:
                    dist = node.x() - phr_node.x()
                    shift = self.fontMetrics().width(node.label) / 2.0 + self.fontMetrics().width(phr_node.label) / 2.0 - abs(dist)
                    if shift > 0:
                        if dist > 0: phr_node.setX(phr_node.x() - shift)
                        else: phr_node.setX(phr_node.x() + shift)
        self.nodes = nodes

def parseTree(lines, hybrid_format):
    if not hybrid_format:
        newlines = []
        for line in lines:
            newlines.append(line.replace('\t-1\t', '\t10000\t'))
        newlines.append('\t'.join(['10000', '[root]', '-1', 'd']))
        return newlines
    else:
        return lines

def initGUI(trees):

    application = QApplication(sys.argv)
    application.setApplicationName("Tree Viewer")
#    application.setApplicationVersion("1.0")
    application.setOrganizationName("Natural Language Processing Centre, Faculty of Informatics, Masaryk University, Brno, Czech Republic")
    application.setOrganizationDomain("http://nlp.fi.muni.cz/trac/set")

    TVForm = TreeViewerForm() # Main Window
#    TVForm.setWindowTitle(application.applicationName() + " " + application.applicationVersion())

    # QSlider shortcuts (cannot be handled in Qt Designer)
    hSliderLeftShortcut = QShortcut(QKeySequence("X"), TVForm)
    hSliderRightShortcut = QShortcut(QKeySequence("C"), TVForm)
    vSliderLeftShortcut = QShortcut(QKeySequence("V"), TVForm)
    vSliderRightShortcut = QShortcut(QKeySequence("B"), TVForm)
    scaleSliderLeftShortcut = QShortcut(QKeySequence("N"), TVForm)
    scaleSliderRightShortcut = QShortcut(QKeySequence("M"), TVForm)
    QObject.connect(hSliderLeftShortcut, SIGNAL("activated()"), TVForm.hSliderSub)
    QObject.connect(hSliderRightShortcut, SIGNAL("activated()"), TVForm.hSliderAdd)
    QObject.connect(vSliderLeftShortcut, SIGNAL("activated()"), TVForm.vSliderSub)
    QObject.connect(vSliderRightShortcut, SIGNAL("activated()"), TVForm.vSliderAdd)
    QObject.connect(scaleSliderLeftShortcut, SIGNAL("activated()"), TVForm.scaleSliderSub)
    QObject.connect(scaleSliderRightShortcut, SIGNAL("activated()"), TVForm.scaleSliderAdd)
    # Status bar
    TVForm.ui.statusbar.showMessage("Use Left/Right arrows or Page Up/Down for switching trees, 'F' for changing the font, 'P' for printing, 'E' for exporting, F5 for restoring default view, 'X'/'C', 'V'/'B', 'N'/'M' for changing the horizontal and vertical tree size and zooming.")
    # Export formats
    TVForm.ui.imageTypes.addItem("SVG")
    TVForm.ui.imageTypes.addItems(map(str, QImageWriter.supportedImageFormats()))
    TVForm.ui.imageTypes.setCurrentIndex(0)

    TVForm.ui.treeview.trees = trees
    TVForm.ui.treeNumber.setMaximum(len(trees) - 1)
    TVForm.ui.treeCount.setText("of " + str(len(trees) - 1))

    TVForm.setTree(0)
    TVForm.show()

    sys.exit(application.exec_())

if __name__ == "__main__":

    trees = []
    treeStart = 0
    hybrid_format = False
    lines = sys.stdin.readlines()
    for lineNumber, line in enumerate(lines):
        lineSplit = line.split()
        if '<top>' in lineSplit or '<TOP>' in lineSplit:
            hybrid_format = True
        if line.isspace() or len(line) == 0 or len(lines) == lineNumber + 1:
            if len(lines) == lineNumber + 1:
                lineNumber += 1
            trees.append(parseTree(lines[treeStart:lineNumber], hybrid_format))
            treeStart = lineNumber + 1
            hybrid_format = False
    initGUI(trees)


