#!/usr/bin/env python

# Copyright 2008 Vojtech Kovar

# This file is part of SET.
#
# SET is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SET.  If not, see <http://www.gnu.org/licenses/>.


# parser itself -- class Parser

import grammar, copy
from matcher import Match, Matcher, match_token
from grammar import Grammar
from segment import Segment, top_token
from token import SurfaceToken, PhrToken, LinkToken
from collocations import Collocations
from utils import log

from settings import COLL_BONUS, VALENCY_BONUS, SMALL_PENALTY, CROSS_PENALTY


BIG_INT = 1000


def is_dep_path(start, target):
		# helps to build an acyclic graph of dependencies
	if start == target: return True
	if start.dependency == target: return True
	curr_token = start
	while curr_token.dependency:
		curr_token = curr_token.dependency
		if curr_token == target: return True
	return False


def is_dep_cross(segment, src_token, dep, cross):
	token = src_token
	while isinstance(token, PhrToken): token = token.head
	if not isinstance(dep, SurfaceToken) or not isinstance(cross, SurfaceToken):
		return False
	node_i = segment.tokens.index(token)
	dep_i = segment.tokens.index(dep)
	cross_i = segment.tokens.index(cross)
	if node_i < cross_i and dep_i > cross_i: return True
	if node_i > cross_i and dep_i < cross_i: return True
	return False


class Parser:

	lw_collx = None
	ll_collx = None
	lt_collx = None
		# for ranking functions
	join_tmpl  = { 'word not': set(['\xc4\x8di']),
						'tag': [{'x': 'S', 'k': '8'}, {'y': 'R', 'k': '3'},
						{'x': 'Q', 'k': '3'}] }
	conj_tmpl  = { 'tag': [{'k': '8'}] }
	verb_tmpl  = { 'tag': [{'k': '5', 'm': 'IBRAPN'}] }
	noun1_tmpl = { 'tag': [{'k': '1', 'c': '1'}] }
	adj1_tmpl  = { 'tag': [{'k': '2', 'c': '1'}] }
	comma_tmpl  = { 'word': [',', '-'] }


	def __init__(self, grammar, lw_collx_root='', ll_collx_root='',
																					lt_collx_root=''):
		if not isinstance(grammar, Grammar):
			raise TypeError('ParserError: Parser init argument must be a Grammar')
		if lw_collx_root:
			lw_dict = lw_collx_root + '.dict'
			lw_file = lw_collx_root + '.collx'
			self.lw_collx = Collocations(lw_dict, lw_file)
		if ll_collx_root:
			ll_dict = ll_collx_root + '.dict'
			ll_file = ll_collx_root + '.collx'
			self.ll_collx = Collocations(ll_dict, ll_file)
		if lt_collx_root:
			lt_dict = lt_collx_root + '.dict'
			lt_file = lt_collx_root + '.collx'
			self.lt_collx = Collocations(lt_dict, lt_file)
		self.grammar = grammar


	def get_collx_bonus(self, match, category=''):
		for i, t1 in enumerate(match.tokens):
			for t2 in match.tokens[i+1:]:
				if category == 'lw':
					score = max( self.lw_collx.find_collocation (
															t1.lemma, set([t2.word])),
									 self.lw_collx.find_collocation (
															set([t1.word]), t2.lemma) )
				elif category == 'll':
					score = self.ll_collx.find_collocation(t1.lemma, t2.lemma)
#					print t1.lemma, t2.lemma, score
				elif category == 'lt':
					try:
						score = 0
						if '5' in t1.tag['k'] and '1' in t2.tag['k']:
							score = self.lt_collx.find_collocation(
											t1.lemma, set(['c' + x for x in t2.tag['c']]))
						elif '5' in t2.tag['k'] and '1' in t1.tag['k']:
							score = self.lt_collx.find_collocation(
											set(['c' + x for x in t1.tag['c']]), t1.lemma)
#						print ['c' + x for x in t1.tag['c']], t1.lemma, score
					except KeyError:
						continue
#						print t1.tag
#						print t2.tag
#						exit(1)
				if score:
					log('Collocations found (%s): %s, %s (%i)'
														% (category, t1.word, t2.word, score))
					return COLL_BONUS * score
		return 1


	def parse(self, segment):
			# method performing complex segment parsing -- adds relationships
			# to the segment
		log('\nParsing segment ' + segment.id + ': '
											+ segment.str_words() + '\n--------------')
			# warning: almost all of the following methods call back the 'parse'
			# method on their subsegments
		self.parse_interjections(segment)
		self.parse_starts(segment)
		self.parse_ends(segment)
		self.parse_hard_delimiters(segment)
		self.parse_relative_clauses(segment)
		self.parse_customs_coords(segment)
		self.parse_dependencies(segment)
		self.select_segment_head(segment)
		log('=======================')


	def parse_interjections(self, segment):
		log('interjections ...')
		matches = self.find_interjections(segment)
		sel_matches = self.select_interjections(segment, matches)
		new_elems = self.make_interjections(segment, sel_matches)
		for elem in new_elems:
			elem.id = segment.id + '.' + str(segment.elements.index(elem))
			self.parse(elem) # parse recursively
		if new_elems: log('\nBack to segment ' + segment.id + ': '
											+ segment.str_words() + '\n--------------')


	def parse_starts(self, segment):
		log('starts ...')
		new_elems = self.make_starts(segment) # this is simple
		for elem in new_elems:
			elem.id = segment.id + '.' + str(segment.elements.index(elem))
			self.parse(elem)
		if new_elems: log('\nBack to segment ' + segment.id + ': '
											+ segment.str_words() + '\n--------------')


	def parse_ends(self, segment):
		log('ends ...')
		new_elems = self.make_ends(segment) # this is simple
		for elem in new_elems:
			elem.id = segment.id + '.' + str(segment.elements.index(elem))
			self.parse(elem)
		if new_elems: log('\nBack to segment ' + segment.id + ': '
											+ segment.str_words() + '\n--------------')


	def parse_hard_delimiters(self, segment):
		log('hard delimiters ...')
		new_elems = self.make_hard_delimiters(segment) # this is also simple
		for elem in new_elems:
			elem.id = segment.id + '.' + str(segment.elements.index(elem))
			self.parse(elem)
		if new_elems: log('\nBack to segment ' + segment.id + ': '
											+ segment.str_words() + '\n--------------')


	def parse_relative_clauses(self, segment):
		while True:
			log('relative clauses ...')
			matches = self.find_relative_clauses(segment)
			sel_matches = self.select_relative_clauses(segment, matches)
			new_elems = self.make_relative_clauses(segment, sel_matches)
			if not new_elems: break
			for elem in new_elems:
				elem.id = segment.id + '.' + str(segment.elements.index(elem))
				self.parse(elem) # parse recursively
			if new_elems:
				log('\nBack to segment ' + segment.id + ': '
												+ segment.str_words() + '\n--------------')


	def parse_customs_coords(self, segment):
		log('coordinations and other constructs ...')
		customs_matches, coords_matches = self.find_customs_coords(segment)
		sel_matches, groups = self.select_customs_coords(segment,
																customs_matches, coords_matches)
		self.make_customs_coords(segment, sel_matches, groups)


	def parse_dependencies(self, segment):
		log('dependencies ...')
		matches = self.find_dependencies(segment)
		sel_matches = self.select_dependencies(segment, matches)
		self.make_dependencies(segment, sel_matches)


	def find_interjections(self, segment):
		# finds all possible interjections in the segment
		matches = []
		for first, second in self.grammar.interjections:
			found_first = []
			found_second = []
			for index, token in enumerate(segment.tokens):
				if token.str_word() == first: found_first.append((index, token))
				if token.str_word() == second: found_second.append((index, token))
			for index1, first in found_first:
				for index2, second in found_second:
					if index1 < index2 and not first.dependency and \
							not second.dependency:
						match = Match(segment, [first, second], index1, index2)
						matches.append(match)
						log('  Match found: ' + match.view())
		return matches
		

	def select_interjections(self, segment, matches):
		# selects best interjections
		for m in matches: m.conflicts = 0
		for index, m1 in enumerate(matches):
			m1.span = m1.index_last - m1.index_first
			for m2 in matches[index+1:]:
				if not m1.index_first > m2.index_last \
						and not m2.index_first > m1.index_last: # conflicting matches
					m1.conflicts -= 1
					m2.conflicts -= 1
		matches_to_sort = [(m.conflicts, m.span, m) for m in matches]
		matches_to_sort.sort(reverse=1)
		selected_matches = []
		for match_triple in matches_to_sort:
			m1 = match_triple[2]
			append = True
			for m2 in selected_matches:
				if not m1.index_first > m2.index_last and \
						not m2.index_first > m1.index_last: # non conflicting matches
					append = False
			if append:
				selected_matches.append(m1)
				log('  Match selected: ' + m1.view())
		return selected_matches
		

	def make_interjections(self, segment, matches):
		# changes segment according to selected matches
		# returns newly added elements
		for match in matches:
			index1 = segment.tokens.index(match.tokens[0])
			index2 = segment.tokens.index(match.tokens[-1])
				# create and add a new subsegment
#			segment.create_elem(index1 + 1, index2, '<sentence>')
				# TODO use this method in the future
			new_segment = Segment( segment.tokens[index1 + 1 : index2] )
			for token in new_segment.tokens:
				dep = token.dependency
				if dep and dep not in new_segment.tokens:
					new_segment.tokens.append(dep)
					del segment.tokens[segment.tokens.index(dep)]
			segment.elements.append(new_segment)
			log('  Sub-segment created: ' + new_segment.str_words())
				# replace subsegment tokens by a link
			segment.tokens[index1 + 1 : index2] = [ LinkToken(
												segment, new_segment, label='<sentence>') ]

				# create a phrasal token for interjection
			index2 = segment.tokens.index(match.tokens[-1]) # actualize index
			children = segment.tokens[index1 : index2 + 1]
			segment.add_phrasal_token( children, '<interjection>',
												segment.tokens[index1 + 1] )
		return segment.elements


	def make_ends(self, segment):
		# returns newly added elements
		if len(segment.tokens) < 2: return []
			#find candidate
		i = len(segment.tokens) - 1
		while True:
			cand_token = segment.tokens[i]
			if isinstance(cand_token, SurfaceToken) or not i: break
			else: i -= 1
		cand = cand_token.str_word()
		for exp in self.grammar.ends:
			if exp.match(cand):
					# create new subsegments
				log('  Match found: ' + cand)
				new_segment1 = Segment(segment.tokens[:i] + segment.tokens[i+1:])
				new_segment2 = Segment([segment.tokens[i]])
				segment.elements.append(new_segment1)
				segment.elements.append(new_segment2)
				log('  Sub-segment created: ' + new_segment1.str_words())
				log('  Sub-segment created: ' + new_segment2.str_words())
					# replace subsegments by links
				segment.tokens = [ 
						LinkToken(segment, new_segment1, label='<sentence>'),
						LinkToken(segment, new_segment2,	label='<ends>') ]
					# create 'top' phrasal token
				segment.add_phrasal_token(copy.copy(segment.tokens), '<top>',
													segment.tokens[0])
				return [new_segment1, new_segment2]
			# no ends
		if '.' not in segment.id:
			new_segment = Segment(segment.tokens)
			segment.elements.append(new_segment)
			log('  Sub-segment created: ' + new_segment.str_words())
			segment.tokens = [LinkToken(segment, new_segment, label='<sentence>')]
			segment.add_phrasal_token(copy.copy(segment.tokens), '<top>',
												segment.tokens[0])
			return [new_segment]
		return []


	def make_starts(self, segment):
		# returns newly added elements
		if len(segment.tokens) < 2: return []
		cand = segment.tokens[0].str_word()
		for exp in self.grammar.starts:
			if exp.match(cand):
					# create new subsegments
				log('  Match found: ' + cand)
				new_segment1 = Segment([segment.tokens[0]])
				new_segment2 = Segment(segment.tokens[1:])
				segment.elements.append(new_segment1)
				segment.elements.append(new_segment2)
				log('  Sub-segment created: ' + new_segment1.str_words())
				log('  Sub-segment created: ' + new_segment2.str_words())
					# replace subsegments by links
				segment.tokens = [ 
						LinkToken(segment, new_segment1, label='<starts>'),
						LinkToken(segment, new_segment2,	label='<sentence>') ]
					# create 'top' phrasal token
				segment.add_phrasal_token(copy.copy(segment.tokens), '<top>',
													segment.tokens[0])
				return [new_segment1, new_segment2]
		return []


	def make_hard_delimiters(self, segment):
		tokens = segment.tokens
		if len(tokens) < 3: return []
		delims = []
		new_segments = []
		for token in tokens[1:-1]:
			if not isinstance(token, SurfaceToken): continue
			if token.word in self.grammar.hard_delimiters:
				delims.append(token)
		if not delims: return [] # no hard delimiter
		for i, delim in enumerate(delims):
			j = tokens.index(delim)
				# create new subsegment
			log('  Match found: ' + delim.str_word())
			new_segment = Segment(tokens[2*i:j])
				# move non-surface tokens
			for token in new_segment.tokens:
				dep = token.dependency
				if dep and dep not in new_segment.tokens:
					new_segment.tokens.append(dep)
					del segment.tokens[tokens.index(dep)]
			new_segments.append(new_segment)
			segment.elements.append(new_segment)
			log('  Sub-segment created: ' + new_segment.str_words())
				# replace subsegment by link
			tokens[2*i:j] = [LinkToken(segment, new_segment, label='<sentence>')]
		# ... and the last one
		i += 1
		new_segment = Segment(tokens[2*i:])
		new_segments.append(new_segment)
		segment.elements.append(new_segment)
		log('  Sub-segment created: ' + new_segment.str_words())
		tokens[2*i:] = [LinkToken(segment, new_segment, label='<sentence>')]
			# create 'top' phrasal token
		segment.add_phrasal_token(copy.copy(tokens), '<sentence>', tokens[1])
		return new_segments


	def find_relative_clauses(self, segment):
		matches = []
		for rule in self.grammar.relative_clauses:
			matcher = Matcher(rule, segment) # skip_matched makes complications
			matches.extend(matcher.get_matches())
		return matches


	def select_relative_clauses(self, segment, matches):
		cands = []
		result = []
			# select candidates
		for iii, m in enumerate(matches):
			good = True; no_joins = 0; no_verbs = 0
			for token in m.all_imp_tokens:
				if match_token(token, self.verb_tmpl): no_verbs += 1
			for token in m.marked_tokens:
				if token.dependency: good = False
			if m.is_imp_closed() and good:
				collx_bonus = 1
#				if self.collx: # correct probabilities
					# ? (not sure if this is good here)
#					collx_bonus =  self.get_collx_bonus(m)
				cands.append((- m.prob, - collx_bonus, m.gap, no_verbs,
																		- m.length, iii, m))
		cands.sort()
		if not cands: return []
			# select matches from candidates
		for prob, collx_bonus, gap, no_verbs, length, iii, m in cands:
			good_one = True
			if m.bound: m_i1 = m.index_first
			else: m_i1 = segment.tokens.index(m.marked_tokens[0])
			if m.rbound: m_i2 = m.index_last
			else: m_i2 = segment.tokens.index(m.marked_tokens[-1])
			for m2 in matches:
				if m2.prob < m.prob: continue
				foreign_dep = segment.tokens.index(m2.dep)
				if m_i1 < foreign_dep and foreign_dep < m_i2:
					good_one = False
			if good_one:
				log('  Match selected: ' + m.view())
				result.append(m)
				return result # want only one of them
		sel_match = cands[0][-1]
		log('  Match selected: ' + sel_match.view());
		return [sel_match]



	def make_relative_clauses(self, segment, matches):
		new_segments = []
		for match in matches:
			if match.bound: index1 = match.get_index_first()
			else: index1 = segment.tokens.index(match.imp_tokens[0]) + 1
			if match.rbound:
				index2 = match.get_index_last() + 1
			else:
				index2 = segment.tokens.index(match.imp_tokens[-1])
				match.marked_tokens = match.marked_tokens[:-1]
				# create and add a new subsegment
			new_segment = Segment( segment.tokens[index1 : index2] )
			for token in new_segment.tokens:
				dep = token.dependency
				if dep and dep not in new_segment.tokens:
					new_segment.tokens.append(dep)
#					print '++++', token.str_word(), dep.str_word()
					del segment.tokens[segment.tokens.index(dep)]
			new_segments.append(new_segment)
			segment.elements.append(new_segment)
			log('  Sub-segment created: ' + new_segment.str_words())
				# replace subsegment tokens by a link
			new_link = LinkToken(segment, new_segment, label='<clause>')
			segment.tokens[index1 : index2] = [ new_link ]
				# create a phrasal token for relclause
			children = match.marked_tokens + [new_link] # TODO ordering ?
			segment.add_phrasal_token( children, '<relclause>',
												new_link, match.dep)
		return new_segments


	def find_customs_coords(self, segment):
		customs_matches = []; coords_matches = []
		for rule in self.grammar.customs:
			matcher = Matcher(rule, segment, skip_matched=True)
			customs_matches.extend(matcher.get_matches())
		for rule in self.grammar.coordinations:
			matcher = Matcher(rule, segment, skip_matched=True)
			coords_matches.extend(matcher.get_matches())
		return customs_matches, coords_matches


	def select_customs_coords(self, segment, customs, coords):
			# rank matches

		def is_correct_multilayer(m1, m2, intersection, coords, customs):
			if len(intersection) > 1: return False
			if m1 in customs and m2 in customs: return False
			token = intersection[0] # the common token
			if m1 in coords and m1.marked_tokens[0] != token \
															and m1.marked_tokens[-1] != token:
				return False
			if m2 in coords and m2.marked_tokens[0] != token \
															and m2.marked_tokens[-1] != token:
				return False
			return True

		iii = 0
		sel_matches = []; ranked_matches = []; groups = []
		for match in coords: # ranking for coords
			iii += 1
			rank = 1.0 * match.prob / (match.imp_length * match.length)
#			if self.collx: # fix prob according to collx
#				rank *= self.get_collx_bonus(match)
			ranked_matches.append((rank, match.index_first, - iii, match))
		correct_chains = True
		while correct_chains:
			ranked_matches.sort(reverse=1)
			correct_chains = False
			for i1, (r1, lindex1, iii1, m1) in enumerate(ranked_matches):
					# fix chain coords ranking
				if r1 <= -BIG_INT: continue
				for i2, (r2, lindex2, iii2, m2) in enumerate(ranked_matches):
					if i1 == i2: continue
					intersection = [tok for tok in m1.marked_tokens
																	if tok in m2.marked_tokens]
					if m2.marked_tokens[0] == m1.marked_tokens[-1] \
										and m1 in coords and m2 in coords: # chain coords
						if r1 <= 0 and r2 > 1:
							ranked_matches[i1] = (r2 - 1, lindex1, iii1, m1)
							correct_chains = True
					elif intersection and r2 > r1: # conflict with better
						log('DELETED ' + m1.view())
						log('BECAUSE ' + m2.view())
						ranked_matches[i1] = (- BIG_INT, lindex1, iii1, m1); break
		for match in customs: # ranking for customs
			iii += 1
			rank = 1.0 * match.prob / (match.imp_length * match.length)
#			if self.collx: # fix prob according to collx
#				rank *= self.get_collx_bonus(match)
			ranked_matches.append((rank, - match.index_first, - iii, match))
		ranked_matches.sort(reverse=1)
			# select compatible
		for rank, lindex, iii, match in ranked_matches:
#			print rank, match.view()
			conflict = -1 # ok
			for i in range(len(sel_matches)):
				sel_m = sel_matches[i]
				intersection = [tok for tok in sel_m.marked_tokens
																if tok in match.marked_tokens]
				if intersection:
						# non-compatible matches unless chain coordinations ...
					if sel_m.marked_tokens[0] == match.marked_tokens[-1] and \
														sel_m in coords and match in coords:
						conflict = i; g_inserted = False # merge
						for g in groups: 
							if g[0] == sel_m: g.insert(0, match); g_inserted = True;
						if not g_inserted: groups.append([match, sel_m])
					elif is_correct_multilayer(match, sel_m, intersection,
																					coords, customs):
						pass  # ... or more phrasal layers
					else:
						conflict = -2; break # conflict
			if conflict > -2 and rank > 0:
				sel_matches.append(match)
		return sel_matches, groups


	def make_customs_coords(self, segment, sel_matches, groups):
			# in 'groups', there can be also matches not present in 'sel_matches'
			# first handle groups of matches ...
		group_matches = []
		for m_list in groups:
			good_group = True; phr_tokens = []
			for match in m_list:
				if not match in sel_matches: good_group = False; break
				if phr_tokens: phr_tokens.extend(match.marked_tokens[1:])
				else: phr_tokens.extend(match.marked_tokens)
				group_matches.append(match)
			if not good_group: continue
			for match in m_list: # logging
				log('  Match selected: ' + match.view() +  ' :: ' + match.label
							+ ' :: ' + ' '.join([tok.str_word() 
														for tok in match.marked_tokens]))
			boss = m_list[-1]
			segment.add_phrasal_token(phr_tokens, boss.label, boss.head, boss.dep)
			# ... and then the rest of matches
		for m in sel_matches:
			if m in group_matches or m.prob == 0: continue
			log('  Match selected: ' + m.view() +  ' :: ' + m.label
						+ ' :: ' + ' '.join([tok.str_word() 
													for tok in m.marked_tokens]))
			segment.add_phrasal_token(m.marked_tokens, m.label, m.head, m.dep)


	def find_dependencies(self, segment):
		matches = []
		for rule in self.grammar.dependencies:
			matcher = Matcher(rule, segment)
			matches.extend(matcher.get_matches())
		for match in matches: # creating depenndencies for phrasal tokens
			# dependency of a phrasal token
			if not match.marked_tokens: continue
			token = match.marked_tokens[0]
			if token.dependency and token.phrasal \
								and token.dependency != match.dep.under_phrase:
				match.prob /= SMALL_PENALTY
				match.marked_tokens[0] = top_token(match.marked_tokens[0])
			# dependency _on_ a phrasal token (not good)
#			if match.marked_tokens and match.dep.dependency and match.dep.phrasal:
#				new_match = copy.copy(match)
#				new_match.dep = match.dep.dependency
#				new_one = True
#				for nm in new_matches:
#					if nm.dep == new_match.dep \
#								and nm.marked_tokens[0] == new_match.marked_tokens[0]:
#						nm.prob = max(nm.prob * SMALL_PENALTY, new_match.prob)
#						nm.length = min (nm.length, new_match.length)
#						new_one = False
#				if new_one:
#					new_match.prob = match.prob / SMALL_PENALTY
#					new_matches.append(new_match)
		return matches


	def select_dependencies(self, segment, matches):
		token_map = {}; ranked_matches = []; lower = {}; sel_matches = []
			# rank matches
		for index, m in enumerate(matches):
			rank = 1.0 * m.prob / m.length
			if m.marked_tokens: token = m.marked_tokens[0]
			else: log('WARNING: no marked tokens: ' + m.view()); continue
			# penalize (some) non-projective dependencies
			top = m.dep
			while top.dependency:
				cross_detect = False
				for child in top.dependency.children:
#					print '+++', m.dep.dependency.str_word(), m.dep.str_word(), token.str_word()
#					print m.view()
#					print token.str_word(), m.dep.str_word()
					if is_dep_cross(segment, token, top, child):
						log('  CROSS EDGE DETECTED: ' + m.view())
						rank /= CROSS_PENALTY
						cross_detect = True; break
				if cross_detect: break
				top = top.dependency
#			print '+++', m.view()
			if token.under_phrase and token.under_phrase != m.dep.under_phrase:
				rank /= CROSS_PENALTY
			if self.lw_collx: # fix prob according to collx
				rank *= self.get_collx_bonus(m, category='lw')
			if self.ll_collx:
				rank *= self.get_collx_bonus(m, category='ll')
			if self.lt_collx:
				rank *= self.get_collx_bonus(m, category='lt')
			# append match with rank
			ranked_matches.append((rank, - index, m))
		ranked_matches.sort(reverse=1)
		return ranked_matches


	def make_dependencies(self, segment, ranked_matches):
		for rank, index, match in ranked_matches: # check if there are no cycles
			if not match.marked_tokens: continue
			token = match.marked_tokens[0]
			dep = match.dep
			# do not allow more dependencies on one preposition
			if '7' in dep.tag.get('k', []) and dep.children:
#				print '++++', dep.str_word(), '::', [t.str_word() for t in dep.children]
				continue
			if not token.dependency and not is_dep_path(dep, token):
				log('  Match selected: ' + match.view())
				log('    Rule: ' + match.rule.view())
				segment.add_dependency(match.marked_tokens[0], match.dep, match.prob)
		if not segment.tokens: log('WARNING: Empty segment'); return
		if len(segment.tokens) == 1: return
			# complete dependencies (somehow but without cycles)
		for index, token in enumerate(segment.tokens):
			if not token.dependency:
				for dep in segment.tokens:
					if not is_dep_path(dep, token):
						log('  Dependency added: ' + token.str_word() + ' -> ' \
																					+ dep.str_word())
						segment.add_dependency(token, dep, 1)
						break


	def select_segment_head(self, segment):
		log('segment head selection ...')
		if not segment.tokens: return
		if len(segment.tokens) == 1: segment.head = segment.tokens[0]
		ranked_tokens = [(self.head_rank(token), token)
								for token in segment.tokens]
		ranked_tokens.sort(reverse=1)
		segment.head = ranked_tokens[0][1]
		log('  Head selected: ' + ranked_tokens[0][1].str_word())


	def head_rank(self, token):
		result = 0
		if not token.dependency                : result += 1000
#		if match_token(token, self.conj_tmpl)  : result +=  200
		if match_token(token, self.verb_tmpl)  : result +=  100
		if match_token(token, self.noun1_tmpl) : result +=   50
		if match_token(token, self.adj1_tmpl)  : result +=   20
		if token.phrasal: result -= 1000 # part of a phrase
		return result - token.dep_prob










