#!/usr/bin/env python

# Copyright 2008 Vojtech Kovar

# This file is part of SET.
#
# SET is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SET.  If not, see <http://www.gnu.org/licenses/>.


# reading segments from text files, class Segment
# and tree output functions

from token import Token, SurfaceToken, PhrToken, LinkToken
from utils import log
import re, sys, copy

brief_re = re.compile('\s+<[lc]>')
compl_lemma = re.compile('([^-`_]+)[-`_]')


def top_token(token):
	t = token
	while t.dependency: t = t.dependency
	return t


def find_surface_dep_id(tok):
	token = tok
	dep = token.dependency
	if not dep:
		if not token.is_head():
			raise RuntimeError('Non-head token without dependency: %s' 
																				% token.str_word())
		if not token.segment.pointer: return -1
		dep = token.segment.pointer.dependency; token = token.segment.pointer
	while not isinstance(dep, SurfaceToken):
		if not dep or dep.str_word() == '<top>': return -1
		if isinstance(dep, PhrToken):
			if token == dep.head: # up
				if dep.is_head(): token = dep.segment.pointer; dep = token.dependency
				else: token = dep; dep = dep.dependency
			else: token = dep; dep = dep.head # down
		elif isinstance(dep, LinkToken): # always down
			dep = dep.target.head
	if dep == tok: return -1
	return dep.id


# classes for more transparent tree drawing and modifying
class TreeNode():

	def __init__(self, node_dict, position=-1):
		try:
			self.id = node_dict['id']
			self.word = node_dict['word']
			self.dep_id = node_dict['dep_id']
			self.dep_type = node_dict['dep_type']
			self.token = node_dict['token']
		except TypeError:
			self.id, self.word, self.dep_id, self.dep_type = node_dict.split()
			self.id = int(self.id)
			self.dep_id = int(self.dep_id)
			self.token = None
		if position == -1: self.position = self.id
		else: self.position = position
		self.children = []
		self.parent = None


class Tree():

	def __init__(self, lines):
		self.nodes = [TreeNode(line) for line in lines]
		self.first_node = None
		self.link_nodes()
		self.fix_phrasal_positions()

	def flat_sentence(self):
		result = ''
		for node in self.nodes:
			if isinstance(node.token, SurfaceToken):
				result += node.token.str_word() + ' '
		return result

	def get_clauses(self):
		return [(self.get_clause_for_node(node), node) for node in self.nodes
								if node.token.str_word() in ('<clause>', '<sentence>')]

	def get_clause_for_node(self, start_node):
		result = []
		q = [start_node]; i = 0
		while i < len(q):
			curr_node = q[i]
			result.append((curr_node.position, curr_node))
			if i == 0 or curr_node.token.str_word() not in ('<clause>', '<relclause>', '<interjection>', '<sentence>'):
				q.extend(curr_node.children)
			i += 1
		result.sort()
		return [n[1] for n in result]

	def link_nodes(self):
		for node in self.nodes:
			if node.dep_id == -1: self.first_node = node
			parent_id = node.dep_id
			for node2 in self.nodes:
				if node2.id == parent_id:
					node.parent = node2
					node2.children.append(node)

	def destroy_non_projectivity(self):
		for node in self.nodes:
			if not isinstance(node.token, SurfaceToken): continue
			for node2 in self.nodes:
				if not node2.parent: continue
				positions = (node2.position, node2.parent.position)
				p1, p2 = min(positions), max(positions)
				while p1 < node.position and p2 > node.position and \
						isinstance(node2.parent.token, SurfaceToken) and \
						not self.ancestor(node, node2.parent):
					del node2.parent.children[node2.parent.children.index(node2)]
					node2.parent = node2.parent.parent
					node2.parent.children.append(node2)
					positions = (node2.position, node2.parent.position)
					p1, p2 = min(positions), max(positions)

	def ancestor(self, start, target):
		node = start
		while node:
			if target == node: return True
			node = node.parent
		return False

	def fix_phrasal_positions(self):
		q = [self.first_node]
		i = 0
		while i < len(q): q.extend(q[i].children); i += 1
		q.reverse()
		for node in q:
			if not isinstance(node.token, SurfaceToken):
				p1, p2 = node.children[0].position, node.children[-1].position
				node.position = (p1 + p2) / 2.0


	def add_preterminals(self):
		length = len(self.nodes)
		for node in self.nodes:
			if not node.children and len(node.parent.children) > 1:
				new_node = TreeNode('%d\t%s\t-2\tp'
												% (length, node.token.get_nterm()))
					# warning: do not set node.position here
				new_node.parent = node.parent
				i = node.parent.children.index(node)
				node.parent.children[i] = new_node
				new_node.children = [node]
				node.parent = new_node
				self.nodes.append(new_node)
				length += 1


	def canonize(self):
		wanted_nterms = set(['<ABBR>', '<OTHER>', '<ADJP>', '<ADVP>', '<VP>',
									'<CP>', '<PP>', '<NP>', '<PRON>', '<SENTENCE>',
									'<ENDS>', '<TOP>', '<UNKNOWN>'])
		q = [self.first_node]
		i = 0
		while i < len(q): q.extend(q[i].children); i += 1
		q.reverse()
		for node in q:
			if node.word.startswith('<') and node.word.upper() not in wanted_nterms:
				node.word = node.children[0].word
			if node.word.upper() in wanted_nterms:
				node.word = node.word.upper()
		self.fix_phrasal_positions()


	def get_laa_format(self, node=None):
		if not node: node = self.first_node
		if node.children:
			inner_part = ' '.join([	self.get_laa_format(child)
												for child in node.children ])
			return '[%s %s ]' % (node.word, inner_part)
		else:
			return node.word


	def sort_children(self):
		for node in self.nodes:
			tmp = []
			for child in node.children: tmp.append((child.position, child))
			tmp.sort()
			node.children = [n for p, n in tmp]


class Segment:

	def __init__(self, tokens, id='0', head=None, pos_tags=False):
		# tokens can be either a list of Token objects or a list of brief lines
		self.elements = [] # subordinate elements created during the analysis
		self.id = id
		self.head = head
		self.pointer = None # pointer token (in case there is one)
		self.pos_tags = pos_tags
		if not tokens: self.tokens = []
#		try:
		self.tokens = self.read_tokens(tokens)
#		except Exception, detail:
#			raise TypeError('SegmentError: Bad token type: %s' % detail)
		for token in self.tokens:
			token.segment = self


	def read_tokens(self, tokens):
		# tokens either as list of Token object or as brief lines
		# brief lines = word, lemma, tag order
		result = []
		first = True
		for i, token in enumerate(tokens):
			if isinstance(token, Token): return tokens # Token list
			line = brief_re.sub('\t', token).strip()
			if not line: continue
			line_list = line.split('\t') + ['', '', '']
			word, lemma, tag = (line_list[0], line_list[1], line_list[2])
			m = compl_lemma.match(lemma)
			if m: lemma = m.group(1);
			result.append(SurfaceToken(self, word, lemma, tag,
								pos_tags=self.pos_tags, first=first, surface_index=i))
			first = False
		return result


	def get_words(self):
		return [token.str_word() for token in self.tokens]


	def str_words(self):
		return ' '.join(self.get_words())


	def create_elem(self, start, end, linkname):
		# currently not used; TODO make parser use this method
		new_segment = Segment( self.tokens[start:end] )
		log('  Sub-segment created: ' + new_segment.str_words())
			# replace subsegment tokens by a link
		new_link = LinkToken(self, new_segment, label=linkname)
		new_segment.pointer = new_link
		self.tokens[start:end] = [new_link]


	def add_phrasal_token(self, children_src_list, phr_name, phr_src_head,
																							dep=None):
		children_list = []
		for child in children_src_list:
			new_child = top_token(child)
			children_list.append(new_child)
			if child == phr_src_head: phr_head = new_child
		new_token = PhrToken(self, phr_name, children=children_list,
									head=phr_head, dependency=dep)
		if dep: dep.children.append(new_token)
		self.tokens.append(new_token)
		for ch in children_list:
			ch.dependency = new_token
			ch.phrasal = True
		# create info about the phrasal token coverage
		i1 = self.tokens.index(children_list[0])
		i2 = self.tokens.index(children_list[-1]) + 1
		for token in self.tokens[i1 : i2]:
			token.under_phrase = new_token

			#logging
		if new_token.head: head_info = ' ::: head = ' + new_token.head.str_word()
		else: head_info = ' ::: head = None'
		if dep: dep_info = ' ::: dep = ' + dep.str_word()
		else: dep_info = ''
		log('  Phrase created: ' + phr_name + ' ::: ' + 
			' '.join(w for w in [tok.str_word() for tok in children_list]) +
			head_info + dep_info)


	def add_dependency(self, child, parent, prob=100):
		if child.dependency: return
		child.dependency = parent
		child.dep_prob = prob
		parent.children.append(child)


	def set_token_ids(self, start_id=0, dep=False):
		id = start_id
		for token in self.tokens:
			if isinstance(token, LinkToken):
				id = token.target.set_token_ids(id, dep)
			if not dep or isinstance(token, SurfaceToken):
				token.id = id
				id += 1
		return id


	def get_tree(self, for_print=True, first=True):
		"returns the tree lines as a list"
		if first: self.set_token_ids()
		result = []
		for token in self.tokens:
				# init values
			id = token.id; word = token.str_word()
			if token.is_head():
				if token.segment.pointer: dep_id = token.segment.pointer.id
				else: dep_id = -1 # the very top token
			else:
#				print '++++', token.str_word(), token.dependency
				if token.dependency: dep_id = token.dependency.id
				else: dep_id = -2 # error
			if token.phrasal or token.is_head(): dep_type = 'p'
			else: dep_type = 'd'
			if isinstance(token, LinkToken):
				result.extend(token.target.get_tree(for_print, first=False))
			if for_print:
				result.append('\t'.join([ str(id), word, str(dep_id), dep_type ]))
			else:
				result.append({'id': id, 'word': word, 'dep_id': dep_id,
									'dep_type': dep_type, 'token': token})
		return result


	def prepare_phr_tree(self):
		tree = Tree(self.get_tree(for_print=False))
		tree.destroy_non_projectivity()
		length = len(tree.nodes)
		q = [tree.first_node]
		i = 0
		# dependency -> phrasal
		q = [tree.first_node]
		i = 0
		while i < len(q): q.extend(q[i].children); i += 1
		for node in q:
			new_children = []
			dep_children = []
			for child in node.children:
				if child.dep_type == 'p': new_children.append(child)
				else: dep_children.append(child)
			node.children = new_children
			if dep_children:
				new_node = TreeNode('%d\t%s\t-2\tp'
												% (length, node.token.get_nterm()))
					# dep_id has no meaning here (is changing)
					# warning: do not set node.position here
				new_node.parent = node.parent
				i = node.parent.children.index(node)
				node.parent.children[i] = new_node
				new_node.children = dep_children + [node]
				node.parent = new_node
				for child in dep_children:
					child.parent = new_node; child.dep_type = 'p'
				tree.nodes.append(new_node)
				length += 1
		# cosmetics
		tree.add_preterminals()
		tree.canonize()
		tree.sort_children()
		return tree


	def get_phr_tree(self):
		self.set_token_ids()
		tree = self.prepare_phr_tree()
		result = []
		for node in tree.nodes:
			if node == tree.first_node:
				node_string = '%s\t%s\t-1\t%s' % (node.id, node.word, node.dep_type)
			else:
				node_string = '%s\t%s\t%s\t%s' % (node.id, node.word, node.parent.id,
																	node.dep_type)
			result.append((int(node.id), node_string))
			result.sort()
		return [item[1] for item in result]


	def get_dep_tree(self, root=True, first=True):
		if first: self.set_token_ids(dep=True)
		result = []
		for token in self.tokens:
			if isinstance(token, PhrToken): continue
			if isinstance(token, LinkToken):
				result.extend(token.target.get_dep_tree(root, first=False))
			if isinstance(token, SurfaceToken):
				id = token.id; word = token.str_word()
				dep_id = find_surface_dep_id(token)
				if root and dep_id == -1:
					result.append('\t'.join([ str(id), word, '10000', 'd' ]))
				else:
					result.append('\t'.join([ str(id), word, str(dep_id), 'd' ]))
		if root and first: result.append('\t'.join(['10000', '[root]', '-1', 'd']))
		return result


	def print_collx_out_for_node(self, node, file=sys.stdout):
		parent_token = node.token
		for child in node.children:
			token = child.token
			if parent_token.lemma and token.lemma:
				for l1 in parent_token.lemma:
					for l2 in token.lemma:
						file.write(l1 + ' <- ' + l2 + '\t(')
						file.write(parent_token.str_tag() + ', ' + token.str_tag())
						file.write(')\t(' + parent_token.str_word())
						file.write(' <- ' + token.str_word() + ')\n')
			self.print_collx_out_for_node(child, file)


	def print_phrases_for_node(self, node, file=sys.stdout, clause=None,
																						marx=False):
			# clause -- print phrases only within a scope of a clause
		if not marx and not node.children: return
		token = node.token
		marx_head = token.tag.get('k', set()).intersection('123467')
		half_head = token.str_word() == '<coord>'
		if not marx or not marx_head:
			for child in node.children:
				self.print_phrases_for_node(child, file, clause, marx)
		if token.str_word() in ['<top>', '<ends>', '<sentence>']: return
		if marx and not marx_head and not half_head: return
		if clause and node not in clause: return
		q = [(node.position, node)]; i = 0
		while i < len(q):
			q.extend([(ch.position, ch) for ch in q[i][1].children
														if not clause or ch in clause])
			i += 1
		q.sort()
		if isinstance(token, SurfaceToken):
			file.write('<phr> (%s): ' % token.str_tag())
		else:
			file.write('%s (%s): ' % (token.str_word(), token.str_tag()))
		if marx:
			file.write(' '.join([n[1].token.str_word() for n in q
													if isinstance(n[1].token, SurfaceToken)]))
			if isinstance(token, SurfaceToken):
				file.write('\n<phrnum> (%s): ' % token.str_tag())
			else:
				file.write('\n%snum> (%s): '
											% (token.str_word()[:-1], token.str_tag()))
			file.write(' '.join([str(n[1].token.surface_index + 1) for n in q
													if isinstance(n[1].token, SurfaceToken)]))
		else:
			file.write(' '.join([n[1].token.str_word() for n in q
													if isinstance(n[1].token, SurfaceToken)]))
		file.write('\n')


	def print_collx_out(self, file=sys.stdout):
		"prints collocations to file"
		tree = Tree(self.get_tree(for_print=False))
		self.print_collx_out_for_node(tree.first_node, file)


	def print_phrases(self, file=sys.stdout):
		"prints phrases to file"
		tree = Tree(self.get_tree(for_print=False))
		self.print_phrases_for_node(tree.first_node, file)


	def print_marx_phrases(self, file=sys.stdout):
		"prints 'marx' phrases to file"
		tree = Tree(self.get_tree(for_print=False))
		clauses = tree.get_clauses()
		file.write('<s>%s\n' % tree.flat_sentence())
		for clause, head_node in clauses: # clause = list of (some) 'tree' nodes
			str_clause = ' '.join([node.token.str_word() for node in clause
											if not node.token.str_word().startswith('<')])
			file.write('<clause>%s\n' % str_clause)
			verb_phrase = ' '.join([node.token.str_word() for node in clause
												if '5' in node.token.tag.get('k', '')])
			file.write('<vp>%s\n' % verb_phrase)
			self.print_phrases_for_node(head_node, file, clause, True)
		file.write('</s>\n')


	def print_laa_tree(self, file=sys.stdout):
		"prints phrasal tree in LAA format to file"
		tree = self.prepare_phr_tree()
		file.write(tree.get_laa_format() + '\n')
		

	def print_tree(self, file=sys.stdout):
		"prints hybrid tree to file"
		file.write('\n'.join(self.get_tree()) + '\n')


	def print_dep_tree(self, file=sys.stdout, root=False):
		"prints dependency tree to file"
		file.write('\n'.join(self.get_dep_tree(root=root)) + '\n')


	def print_phr_tree(self, file=sys.stdout):
		"prints phrasal tree to file"
		file.write('\n'.join(self.get_phr_tree()) + '\n')


