#!/usr/bin/env python

# Copyright 2008 Vojtech Kovar

# This file is part of SET.
#
# SET is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SET.  If not, see <http://www.gnu.org/licenses/>.


# parser itself -- class Parser

import grammar, copy
from matcher import Match, Matcher, match_token
from grammar import Grammar
from segment import Segment, top_token
from settoken import SurfaceToken, PhrToken, LinkToken
from collocations import Collocations
from utils import log

from settings import COLL_BONUS, VALENCY_BONUS, SMALL_PENALTY, CROSS_PENALTY


BIG_INT = 1000


def is_dep_path(start, target):
        # helps to build an acyclic graph of dependencies
    if start == target: return True
    if start.dependency == target: return True
    curr_token = start
    while curr_token.dependency:
        curr_token = curr_token.dependency
        if curr_token == target: return True
    return False


def is_dep_cross(segment, src_token, dep, cross):
    token = src_token
    while isinstance(token, PhrToken): token = token.head
    if not isinstance(dep, SurfaceToken) or not isinstance(cross, SurfaceToken):
        return False
    node_i = segment.tokens.index(token)
    dep_i = segment.tokens.index(dep)
    cross_i = segment.tokens.index(cross)
    if node_i < cross_i and dep_i > cross_i: return True
    if node_i > cross_i and dep_i < cross_i: return True
    return False


class Parser:

    lw_collx = None
    ll_collx = None
    lt_collx = None
        # for ranking functions
    join_tmpl  = { 'word not': set(['\xc4\x8di']),
                   'tag': [{'x': 'S', 'k': '8'}, {'y': 'R', 'k': '3'},
                           {'x': 'Q', 'k': '3'}] }
    conj_tmpl  = { 'tag': [{'k': '8'}] }
    verb_tmpl  = { 'tag': [{'k': '5', 'm': 'IBRAPN'}] }
    noun1_tmpl = { 'tag': [{'k': '1', 'c': '1'}] }
    adj1_tmpl  = { 'tag': [{'k': '2', 'c': '1'}] }
    comma_tmpl  = { 'word': [',', '-'] }


    def __init__(self, grammar, lw_collx_root='', ll_collx_root='',
                                                             lt_collx_root=''):
        if not isinstance(grammar, Grammar):
            raise TypeError(
                         'ParserError: Parser init argument must be a Grammar')
        if lw_collx_root:
            lw_dict = lw_collx_root + '.dict'
            lw_file = lw_collx_root + '.collx'
            self.lw_collx = Collocations(lw_dict, lw_file)
        if ll_collx_root:
            ll_dict = ll_collx_root + '.dict'
            ll_file = ll_collx_root + '.collx'
            self.ll_collx = Collocations(ll_dict, ll_file)
        if lt_collx_root:
            lt_dict = lt_collx_root + '.dict'
            lt_file = lt_collx_root + '.collx'
            self.lt_collx = Collocations(lt_dict, lt_file)
        self.grammar = grammar


    def get_collx_bonus(self, match, category=''):
        for i, t1 in enumerate(match.tokens):
            for t2 in match.tokens[i+1:]:
                if category == 'lw':
                    score = max( self.lw_collx.find_collocation (
                                                    t1.lemma, set([t2.word])),
                                 self.lw_collx.find_collocation (
                                                    set([t1.word]), t2.lemma) )
                elif category == 'll':
                    score = self.ll_collx.find_collocation(t1.lemma, t2.lemma)
#                    print t1.lemma, t2.lemma, score
                elif category == 'lt':
                    try:
                        score = 0
                        if '5' in t1.tag['k'] and '1' in t2.tag['k']:
                            score = self.lt_collx.find_collocation(
                                 t1.lemma, set(['c' + x for x in t2.tag['c']]))
                        elif '5' in t2.tag['k'] and '1' in t1.tag['k']:
                            score = self.lt_collx.find_collocation(
                                 set(['c' + x for x in t1.tag['c']]), t1.lemma)
                    except KeyError:
                        continue
                if score:
                    log('Collocations found (%s): %s, %s (%i)'
                                         % (category, t1.word, t2.word, score))
                    return COLL_BONUS * score
        return 1


    def parse(self, segment):
            # method performing complex segment parsing -- adds relationships
            # to the segment
        matches = self.find_matches(segment)
        sorted_matches = self.sort_matches(segment, matches)
        self.make_matches(segment, sorted_matches)
        self.finish_tree(segment)


    def find_matches(self, segment):
        matches = []
        wanted_rules = self.grammar.coordinations + self.grammar.dependencies \
                                + self.grammar.relative_clauses
        for rule in wanted_rules:
            matcher = Matcher(rule, segment)
            matches.extend(matcher.get_matches())
        return matches


    def sort_matches(self, segment, matches):
        ranked_matches = []
        for index, m in enumerate(matches):
            rank = 1.0 * m.prob / m.length # could be improved later
            ranked_matches.append((rank, - index, m))
        ranked_matches.sort(reverse=1)
        return ranked_matches


    def make_matches(self, segment, sorted_matches):
        for rank, index, m in sorted_matches: # m = match
            if not m.marked_tokens:
                log('WARNING: no marked tokens: ' + m.view()); continue
            if m.rule in self.grammar.coordinations:
                if m.imp_tokens and m.imp_tokens[0].dependency \
                        and m.imp_tokens[-1].dependency \
                        and not m.imp_tokens[0].phrasal \
                        and not m.imp_tokens[-1].phrasal:
                    continue # overriden by other bounds
                conflict_tokens = [t for t in m.tokens[1:-1] if t.under_phrase]
                if conflict_tokens: continue
                conflict_tokens = [t for t in m.marked_tokens if t.phrasal]
                conflict = False; new_on_left = False; new_on_right = False
                for c_token in conflict_tokens: # check chains
                    fcp_token = c_token.dependency.children[0]
                                        # first token in the conflict phrase
                    lcp_token = c_token.dependency.children[-1]
                                        # last token in the conflict phrase
                    if c_token == m.marked_tokens[-1] and fcp_token == c_token:
                        new_on_left = True
                    elif c_token == m.marked_tokens[0] and lcp_token == c_token:
                        new_on_right = True
                    else:
                        conflict = True
                if conflict: continue
                dep_right, dep_left = None, None
                if new_on_left: # update the match and delete the old coord
                    c_token = m.marked_tokens[-1]
                    m.marked_tokens.extend(c_token.dependency.children[1:])
                    m.head = c_token.dependency.head
                    dep_left = c_token.dependency.dependency
                    segment.del_phrasal_token(c_token.dependency)
                if new_on_right:
                    c_token = m.marked_tokens[0]
                    m.marked_tokens = c_token.dependency.children[:-1] \
                                                              + m.marked_tokens
                    dep_right = c_token.dependency.dependency
                    segment.del_phrasal_token(c_token.dependency)
                log('  Match selected: %s (rank %s)' % (m.view(), rank))
                log('     Rule: ' + m.rule.view())
                dep = dep_right or dep_left
                if not dep and new_on_left and not new_on_right:
                    dep = m.marked_tokens[0].dependency 
                elif not dep and new_on_right and not new_on_left:
                    dep = m.marked_tokens[-1].dependency
                for t in m.marked_tokens: t.dependency = None
                segment.add_phrasal_token(m.marked_tokens, '<coord>', m.head,
                                                                           dep)
            elif not m.rule in self.grammar.dependencies:
                conflict_tokens = [t for t in m.marked_tokens
                                       if t.under_phrase
                                         or (t.dependency
                                           and t != t.dependency.children[0]
                                           and t != t.dependency.children[-1])]
                if conflict_tokens: continue
                log('  Match selected: %s (rank %s)' % (m.view(), rank))
                log('    Rule: ' + m.rule.view())
                segment.add_phrasal_token(m.marked_tokens, m.label, m.head,
                                                       all_tokens=m.all_tokens)
            else: # dependencies
                if not m.marked_tokens: continue
                token = m.marked_tokens[0]
                dep = m.dep
                # do not allow more dependencies on one preposition
                # (XXX kind of hack)
                if '7' in dep.tag.get('k', []) and dep.children:
                    continue
                if not token.dependency and not is_dep_path(dep, token):
                    log('  Match selected: %s (rank %s)' % (m.view(), rank))
                    log('     Rule: ' + m.rule.view())
                    segment.add_dependency(m.marked_tokens[0], m.dep, m.prob)


    def finish_tree(self, segment):
        # complete dependencies (somehow -- add top phrasal token)
        log('completing the tree ...')
        orphan_tokens = []
        for index, token in enumerate(segment.tokens):
            if not token.dependency:
                if token.under_phrase:
                    segment.add_dependency(token, token.under_phrase[0])
                    token.phrasal = True # XXX kind of hack
                    log('  Added edge: %s -> %s'
                        % (token.str_word(), token.under_phrase[0].str_word()))
                else:
                    orphan_tokens.append(token)
        segment.head = segment.add_phrasal_token(orphan_tokens, '<sentence>',
                                               self.select_head(orphan_tokens))


    def select_head(self, tokens):
        if not tokens: return
        if len(tokens) == 1: return tokens[0]
        ranked_tokens = [(self.head_rank(token), token) for token in tokens]
        ranked_tokens.sort(reverse=1)
        return ranked_tokens[0][1]


    def head_rank(self, token):
        result = 0
        if not token.dependency                : result += 1000
        if match_token(token, self.verb_tmpl)  : result +=  100
        if match_token(token, self.noun1_tmpl) : result +=   50
        if match_token(token, self.adj1_tmpl)  : result +=   20
        if token.phrasal: result -= 1000 # part of a phrase
        return result - token.dep_prob










