#!/usr/bin/env python
# -*- coding: utf8 -*-

# Copyright 2008 Vojtech Kovar

# This file is part of SET.
#
# SET is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SET is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SET.  If not, see <http://www.gnu.org/licenses/>.


# contains basic class -- Token and specialized token classes

import copy, re
import settings


def parse_lemma(str_lemma):
	if len(str_lemma) > 1 and ',' in str_lemma:
		return set(str_lemma.split(','))
	return set([str_lemma])


def tag_to_dict(str_tag, pos_tags=False):
	result = {}
	if pos_tags: tag = translate_pos_tag(str_tag)
	else:
		tag = str_tag
		if not tag.endswith(','):
			tag_list = str_tag.replace(';', '').split(',')
			tag = ''.join([t for t in tag_list if 'wH' not in t])
	# few quick hacks follow ...
	if not pos_tags:
		if 'gU' in tag: tag += 'gMgIgN'
		elif tag == 'Vc-------------': tag = 'k5eAp3nPtPmCa.'
		elif tag.startswith('D'): tag = 'k6'
		elif tag.startswith('V'): tag = 'k5'
		elif tag == 'PHZS2--3-------': tag = 'k3xPc2p3g.nS'
		elif tag.endswith('xxx'): tag = tag [:-3]
	# end hacks
	if not len(tag) % 2 == 0:
		if settings.IGNORE_WRONG_MORPHOLOGY: return {}
		exc_text = 'TokenError: corrupted tag: "%s"' % tag
		if pos_tags:
			exc_text += '(positional tags conversion failure: "%s")' % str_tag
		raise RuntimeError(exc_text)
	for i in range(len(tag) / 2):
		key = tag[2*i]
		if not result.get(key): result[key] = set()
		result[key].add(tag[2*i + 1])
	return result


def translate_pos_tag(str_tag):
	transl_table = {
		1: {  'A': 'k2', 'C': 'k4', 'D': 'k6', 'I': 'k0', 'J': 'k8', 'N': 'k1',
				'P': 'k3', 'V': 'k5', 'R': 'k7', 'T': 'k9', 'X': 'k.', 'Z': 'kI',
			},
		2: {  ',': 'xS', '1': 'xS', '4': 'yR', '5': 'xP', '6': 'xP', '8': 'xO',
				'9': 'yR', '?': 'yR', 'B': 'mI', 'C': 'mN', 'D': 'xD', 'E': 'yR',
				'F': 'q1', 'H': 'xPqC', # q1 = 1st part of prep, qC = clitical
				'J': 'yR', 'K': 'yR', 'L': 'yI', 'O': 'xO',
				'P': 'xP', 'Q': 'yR', 'S': 'xO', 'U': 'yR', 'W': 'yN', 'Y': 'yR',
				'Z': 'xI', '^': 'xC', 'a': 'yI', 'b': 'eAd1', 'c': 'kYmC',
				'd': 'k2', # ?
				'e': 'mS', 'f': 'mF', 'i': 'mR',
				'j': 'qN', 'k': 'qA', # qN = noun, qA = adjective
				'm': 'mD', 'o': 'k6eAd1', 'p': 'mA', 'q': 'mA', 's': 'mN',
				't': 'mI', 'u': 'k6yR', 'v': 'k6eAd1', 'w': 'k2', # ?
				'y': 'k1gF', 'z': 'yR',
				'!': 'k6', '.': 'k2', '3': 'k4', ';': 'k1', 'x': 'kA', '~': 'k5',
			},
		3: {  'F': 'gF', 'H': 'gNgF', 'I': 'gI', 'M': 'gM', 'N': 'gN', 'Q': 'gNgF',
				'T': 'nPgIgF', 'X': 'g.', 'Y': 'gMgI', 'Z': 'gMgNgI',
			},
		4: {  'D': 'nP', 'P': 'nP', 'S': 'nS', 'W': 'n.', 'X': 'n.',
			},
		5: {  '1': 'c1', '2': 'c2', '3': 'c3', '4': 'c4', '5': 'c5', '6': 'c6',
				'7': 'c7', 'X': 'c.',
			},
		6: {}, # not interesting (at least so far)
		7: {},
		8: {  '1': 'p1', '2': 'p2', '3': 'p3', 'X': 'p.',
			},
		9: {  'F': 'mBmI', 'P': 'mI', 'R': 'mR', # 'X': 'm.' ? (m already in [2])
			},
		10:{  '1': 'd1', '2': 'd2', '3': 'd3',
			},
		11:{  'A': 'eA', 'N': 'eN',
			},
		12:{  'P': 'mN', # 'A' is too broad
			},
		13:{},
		14:{},
		15:{},
	}
	result = ''
	for i in range(15):
		result += transl_table[i+1].get(str_tag[i], '')
	return result


class Token:

	def __init__(self, segment, dependency=None, phrasal=False, children=[],
						dep_prob=100, surface_index=-1):
		self.dependency = dependency # token - the superior one
		self.phrasal = phrasal # defines if the 'dependency' link is phrasal
		self.segment = segment # segment containing this token
		self.children = copy.copy(children) # list of child tokens
		self.dep_prob = dep_prob # probability of dependency
		self.word = ''
		self.lemma = set()
		self.tag = {}
		self.under_phrase = None # helps in advanced parts of parsing
		self.surface_index = surface_index


	def get_nterm(self):
		pos = self.tag.get('k', '')
		if not pos: return '<X>'
		if pos.intersection('A') : return '<ABBR>'
		if pos.intersection('90'): return '<OTHER>'
		if pos.intersection('2') : return '<ADJP>'
		if pos.intersection('6') : return '<ADVP>'
		if pos.intersection('5') : return '<VP>'
		if pos.intersection('8I'): return '<CP>'
		if pos.intersection('7') : return '<PP>'
		if pos.intersection('1') : return '<NP>'
		if pos.intersection('3') : return '<PRON>'
		if pos.intersection('4'):
			if self.tag.get('x', set()).intersection('C'): return '<NP>'
			else: return '<ADJP>'
		return '<UNKNOWN>'


	def is_head(self):
		return (self.segment.head == self)


	def str_tag(self):
		result = ''
		for key, value_list in self.tag.iteritems():
			for value in value_list:
				result += key + value
		return result


class SurfaceToken(Token):

	num_re = re.compile('^[0-9][0-9,.]*$')
	punct_re = re.compile('^&.*;$')
	abbr_re = re.compile('^[A-Z][A-Z]+$')
#	upper_re = re.compile('^[A-ZĂÄÄĂĂĹĂĹĹ Ĺ¤ĂĂ]')

	def __init__(self, segment, word, lemma, tag, dependency=None,	phrasal=False,
						children=[], dep_prob=100, surface_index=-1, pos_tags=False,
						first=False):
		Token.__init__(self, segment, dependency, phrasal, children, dep_prob,
							surface_index)
		self.word = word # as string
		self.lemma = parse_lemma(lemma) # as set of strings
		self.tag = tag_to_dict(tag, pos_tags) # as dict
		if not self.tag:
			if self.num_re.match(word):
				self.tag = {'k': set(['4'])} # numeral
				if self.word == '1': self.tag['n'] = set(['S'])
				else: self.tag['n'] = set(['P'])
			if word in '*,.%=-_;:!?/\'"()+|[]' or self.punct_re.match(word):
				self.tag = {'k': set(['I'])} # punctuation
#		print word, lemma, tag, self.tag
		if not self.tag.get('k', ''): self.tag['k'] = set()
		if not self.tag.get('m', ''): self.tag['m'] = set()
		if self.abbr_re.match(word):
			self.tag['k'].add('A')
		if unicode.isupper(unicode(word, 'utf8')[0]) and not first:
			self.tag['k'].add('1')
		if word == "lze":
			self.tag['k'].add('5')
			self.tag['m'] = set(['I'])
		if word == "jako":
			self.tag['k'].add('7')

	def str_word(self):
		return self.word


class PhrToken(Token):

	def __init__(self, segment, label,  tag='', dependency=None, phrasal=False,
						children=[], head=None, dep_prob=100):
		Token.__init__(self, segment, dependency, phrasal, children, dep_prob, -1)
		self.label = label # as string
		self.tag = tag_to_dict(tag) # as dict
		self.head = head # token
			# inherit head attributes
# THIS IS POSSIBLE ONLY IF THE TOKEN HAS THE RIGHT POSITION IN THE SENTENCE
#		if head:
#			self.word = head.word
#			self.lemma = head.lemma
#			self.tag = head.tag


	def str_word(self):
		return self.label


class LinkToken(Token):

	def __init__(self, segment, target, dependency=None, phrasal=False,
						children=[], label='', dep_prob=100):
		Token.__init__(self, segment, dependency, phrasal, children, dep_prob, -1)
		self.target = target # segment - target of the link
		self.label = label # does not have to be displayed...
		target.pointer = self


	def str_word(self):
		return self.label

